@@ -29,6 +29,38 @@ use crate::RunCtx;
 
 type Result<T> = result::Result<T, Error>;
 
+/// Wraps a writer and buffers its output.
+///
+/// This implementation is similar to `io::BufWriter`, but:
+/// - The inner writer is only written to if `flush` is called.
+/// - The buffer may grow infinitely large.
+struct BufWriter<'w, W: io::Write + ?Sized> {
+  buf: Vec<u8>,
+  inner: &'w mut W,
+}
+
+impl<'w, W: io::Write + ?Sized> BufWriter<'w, W> {
+  pub fn new(inner: &'w mut W) -> Self {
+    BufWriter {
+      buf: Vec::with_capacity(128),
+      inner,
+    }
+  }
+}
+
+impl<'w, W: io::Write + ?Sized> io::Write for BufWriter<'w, W> {
+  fn write(&mut self, buf: &[u8]) -> io::Result<usize> {
+    self.buf.extend_from_slice(buf);
+    Ok(buf.len())
+  }
+
+  fn flush(&mut self) -> io::Result<()> {
+    self.inner.write_all(&self.buf)?;
+    self.buf.clear();
+    self.inner.flush()
+  }
+}
+
 trait Stdio {
   fn stdio(&mut self) -> (&mut dyn io::Write, &mut dyn io::Write);
 }
@@ -39,6 +71,15 @@ impl<'io> Stdio for RunCtx<'io> {
   }
 }
 
+impl<W> Stdio for (&mut W, &mut W)
+where
+  W: io::Write,
+{
+  fn stdio(&mut self) -> (&mut dyn io::Write, &mut dyn io::Write) {
+    (self.0, self.1)
+  }
+}
+
 /// A command execution context that captures additional data pertaining
 /// the command execution.
 pub struct ExecCtx<'io> {
@@ -50,6 +91,7 @@ pub struct ExecCtx<'io> {
   pub new_admin_pin: Option<ffi::OsString>,
   pub new_user_pin: Option<ffi::OsString>,
   pub password: Option<ffi::OsString>,
+  pub no_cache: bool,
   pub verbosity: u64,
 }
 
@@ -60,9 +102,10 @@ impl<'io> Stdio for ExecCtx<'io> {
 }
 
 /// The available Nitrokey models.
+#[allow(unused_doc_comments)]
 Enum! {DeviceModel, [
   Pro => "pro",
-  Storage => "storage"
+  Storage => "storage",
 ]}
 
 impl From<DeviceModel> for nitrokey::Model {
@@ -75,19 +118,21 @@ impl From<DeviceModel> for nitrokey::Model {
 }
 
 /// A top-level command for nitrocli.
+#[allow(unused_doc_comments)]
 Enum! {Command, [
   Config => ("config", config),
   Lock => ("lock", lock),
   Otp => ("otp", otp),
   Pin => ("pin", pin),
   Pws => ("pws", pws),
+  Reset => ("reset", reset),
   Status => ("status", status),
-  Storage => ("storage", storage)
+  Storage => ("storage", storage),
 ]}
 
 Enum! {ConfigCommand, [
   Get => ("get", config_get),
-  Set => ("set", config_set)
+  Set => ("set", config_set),
 ]}
 
 #[derive(Clone, Copy, Debug)]
@@ -129,17 +174,17 @@ Enum! {OtpCommand, [
   Clear => ("clear", otp_clear),
   Get => ("get", otp_get),
   Set => ("set", otp_set),
-  Status => ("status", otp_status)
+  Status => ("status", otp_status),
 ]}
 
 Enum! {OtpAlgorithm, [
   Hotp => "hotp",
-  Totp => "totp"
+  Totp => "totp",
 ]}
 
 Enum! {OtpMode, [
   SixDigits => "6",
-  EightDigits => "8"
+  EightDigits => "8",
 ]}
 
 impl From<OtpMode> for nitrokey::OtpMode {
@@ -154,49 +199,58 @@ impl From<OtpMode> for nitrokey::OtpMode {
 Enum! {OtpSecretFormat, [
   Ascii => "ascii",
   Base32 => "base32",
-  Hex => "hex"
+  Hex => "hex",
 ]}
 
 Enum! {PinCommand, [
   Clear => ("clear", pin_clear),
   Set => ("set", pin_set),
-  Unblock => ("unblock", pin_unblock)
+  Unblock => ("unblock", pin_unblock),
 ]}
 
 Enum! {PwsCommand, [
   Clear => ("clear", pws_clear),
   Get => ("get", pws_get),
   Set => ("set", pws_set),
-  Status => ("status", pws_status)
+  Status => ("status", pws_status),
 ]}
 
 fn parse(
   ctx: &mut impl Stdio,
-  parser: &argparse::ArgumentParser<'_>,
+  parser: argparse::ArgumentParser<'_>,
   args: Vec<String>,
 ) -> Result<()> {
   let (stdout, stderr) = ctx.stdio();
-  if let Err(err) = parser.parse(args, stdout, stderr) {
-    Err(Error::ArgparseError(err))
-  } else {
-    Ok(())
-  }
+  let result = parser
+    .parse(args, stdout, stderr)
+    .map_err(Error::ArgparseError);
+  drop(parser);
+  result
 }
 
 /// Inquire the status of the nitrokey.
 fn status(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Prints the status of the connected Nitrokey device");
-  parse(ctx, &parser, args)?;
+  parse(ctx, parser, args)?;
 
   commands::status(ctx)
 }
 
+/// Perform a factory reset.
+fn reset(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
+  let mut parser = argparse::ArgumentParser::new();
+  parser.set_description("Performs a factory reset");
+  parse(ctx, parser, args)?;
+
+  commands::reset(ctx)
+}
+
 Enum! {StorageCommand, [
   Close => ("close", storage_close),
   Hidden => ("hidden", storage_hidden),
   Open => ("open", storage_open),
-  Status => ("status", storage_status)
+  Status => ("status", storage_status),
 ]}
 
 /// Execute a storage subcommand.
@@ -217,8 +271,7 @@ fn storage(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
     "The arguments for the subcommand",
   );
   parser.stop_on_first_argument(true);
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
   subargs.insert(0, format!("nitrocli {} {}", Command::Storage, subcommand));
   subcommand.execute(ctx, subargs)
@@ -228,7 +281,7 @@ fn storage(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
 fn storage_open(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Opens the encrypted volume on a Nitrokey Storage");
-  parse(ctx, &parser, args)?;
+  parse(ctx, parser, args)?;
 
   commands::storage_open(ctx)
 }
@@ -237,7 +290,7 @@ fn storage_open(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
 fn storage_close(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Closes the encrypted volume on a Nitrokey Storage");
-  parse(ctx, &parser, args)?;
+  parse(ctx, parser, args)?;
 
   commands::storage_close(ctx)
 }
@@ -246,7 +299,7 @@ fn storage_close(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
 fn storage_status(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Prints the status of the Nitrokey's storage");
-  parse(ctx, &parser, args)?;
+  parse(ctx, parser, args)?;
 
   commands::storage_status(ctx)
 }
@@ -254,7 +307,7 @@ fn storage_status(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
 Enum! {HiddenCommand, [
   Close => ("close", storage_hidden_close),
   Create => ("create", storage_hidden_create),
-  Open => ("open", storage_hidden_open)
+  Open => ("open", storage_hidden_open),
 ]}
 
 /// Execute a storage hidden subcommand.
@@ -275,8 +328,7 @@ fn storage_hidden(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
     "The arguments for the subcommand",
   );
   parser.stop_on_first_argument(true);
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
   subargs.insert(
     0,
@@ -313,8 +365,7 @@ fn storage_hidden_create(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()>
     "The end location of the hidden volume as percentage of the \
      encrypted volume's size (1-100)",
   );
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
   commands::storage_hidden_create(ctx, slot, start, end)
 }
@@ -322,7 +373,7 @@ fn storage_hidden_create(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()>
 fn storage_hidden_open(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Opens a hidden volume on a Nitrokey Storage");
-  parse(ctx, &parser, args)?;
+  parse(ctx, parser, args)?;
 
   commands::storage_hidden_open(ctx)
 }
@@ -330,7 +381,7 @@ fn storage_hidden_open(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
 fn storage_hidden_close(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Closes the hidden volume on a Nitrokey Storage");
-  parse(ctx, &parser, args)?;
+  parse(ctx, parser, args)?;
 
   commands::storage_hidden_close(ctx)
 }
@@ -353,8 +404,7 @@ fn config(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
     "The arguments for the subcommand",
   );
   parser.stop_on_first_argument(true);
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
   subargs.insert(0, format!("nitrocli {} {}", Command::Config, subcommand));
   subcommand.execute(ctx, subargs)
@@ -364,7 +414,7 @@ fn config(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
 fn config_get(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Prints the Nitrokey configuration");
-  parse(ctx, &parser, args)?;
+  parse(ctx, parser, args)?;
 
   commands::config_get(ctx)
 }
@@ -421,8 +471,7 @@ fn config_set(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
     argparse::StoreTrue,
     "Allow one-time password generation without PIN",
   );
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
   let numlock = ConfigOption::try_from(no_numlock, numlock, "numlock")?;
   let capslock = ConfigOption::try_from(no_capslock, capslock, "capslock")?;
@@ -441,7 +490,7 @@ fn config_set(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
 fn lock(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Locks the connected Nitrokey device");
-  parse(ctx, &parser, args)?;
+  parse(ctx, parser, args)?;
 
   commands::lock(ctx)
 }
@@ -464,8 +513,7 @@ fn otp(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
     "The arguments for the subcommand",
   );
   parser.stop_on_first_argument(true);
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
   subargs.insert(0, format!("nitrocli {} {}", Command::Otp, subcommand));
   subcommand.execute(ctx, subargs)
@@ -496,8 +544,7 @@ fn otp_get(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
     argparse::StoreOption,
     "The time to use for TOTP generation (Unix timestamp, default: system time)",
   );
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
   commands::otp_get(ctx, slot, algorithm, time)
 }
@@ -568,8 +615,7 @@ pub fn otp_set(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
     argparse::StoreOption,
     &fmt_help,
   );
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
   if ascii {
     if secret_format.is_some() {
@@ -616,8 +662,7 @@ fn otp_clear(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let _ = parser
     .refer(&mut algorithm)
     .add_option(&["-a", "--algorithm"], argparse::Store, &help);
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
   commands::otp_clear(ctx, slot, algorithm)
 }
@@ -632,8 +677,7 @@ fn otp_status(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
     argparse::StoreTrue,
     "Show slots that are not programmed",
   );
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
   commands::otp_status(ctx, all)
 }
@@ -656,8 +700,7 @@ fn pin(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
     "The arguments for the subcommand",
   );
   parser.stop_on_first_argument(true);
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
   subargs.insert(0, format!("nitrocli {} {}", Command::Pin, subcommand));
   subcommand.execute(ctx, subargs)
@@ -667,7 +710,7 @@ fn pin(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
 fn pin_clear(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Clears the cached PINs");
-  parse(ctx, &parser, args)?;
+  parse(ctx, parser, args)?;
 
   commands::pin_clear(ctx)
 }
@@ -682,8 +725,7 @@ fn pin_set(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
     .refer(&mut pintype)
     .required()
     .add_argument("type", argparse::Store, &help);
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
   commands::pin_set(ctx, pintype)
 }
@@ -692,7 +734,7 @@ fn pin_set(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
 fn pin_unblock(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Unblocks and resets the user PIN");
-  parse(ctx, &parser, args)?;
+  parse(ctx, parser, args)?;
 
   commands::pin_unblock(ctx)
 }
@@ -715,8 +757,7 @@ fn pws(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
     "The arguments for the subcommand",
   );
   parser.stop_on_first_argument(true);
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
   subargs.insert(0, format!("nitrocli {} {}", Command::Pws, subcommand));
   subcommand.execute(ctx, subargs)
@@ -756,8 +797,7 @@ fn pws_get(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
     argparse::StoreTrue,
     "Print the stored data without description",
   );
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
   commands::pws_get(ctx, slot, name, login, password, quiet)
 }
@@ -790,8 +830,7 @@ fn pws_set(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
     argparse::Store,
     "The password to store on the slot",
   );
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
   commands::pws_set(ctx, slot, &name, &login, &password)
 }
@@ -806,8 +845,7 @@ fn pws_clear(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
     argparse::Store,
     "The PWS slot to clear",
   );
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
   commands::pws_clear(ctx, slot)
 }
@@ -822,18 +860,16 @@ fn pws_status(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
     argparse::StoreTrue,
     "Show slots that are not programmed",
   );
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
   commands::pws_status(ctx, all)
 }
 
-/// Parse the command-line arguments and return the selected command and
-/// the remaining arguments for the command.
-fn parse_arguments<'io, 'ctx: 'io>(
-  ctx: &'ctx mut RunCtx<'_>,
-  args: Vec<String>,
-) -> Result<(Command, ExecCtx<'io>, Vec<String>)> {
+/// Parse the command-line arguments and execute the selected command.
+pub(crate) fn handle_arguments(ctx: &mut RunCtx<'_>, args: Vec<String>) -> Result<()> {
+  use std::io::Write;
+
+  let mut version = false;
   let mut model: Option<DeviceModel> = None;
   let model_help = format!(
     "Select the device model to connect to ({})",
@@ -844,6 +880,11 @@ fn parse_arguments<'io, 'ctx: 'io>(
   let cmd_help = cmd_help!(command);
   let mut subargs = vec![];
   let mut parser = argparse::ArgumentParser::new();
+  let _ = parser.refer(&mut version).add_option(
+    &["-V", "--version"],
+    argparse::StoreTrue,
+    "Print version information and exit",
+  );
   let _ = parser.refer(&mut verbosity).add_option(
     &["-v", "--verbose"],
     argparse::IncrBy::<u64>(1),
@@ -864,27 +905,34 @@ fn parse_arguments<'io, 'ctx: 'io>(
     "The arguments for the command",
   );
   parser.stop_on_first_argument(true);
-  parse(ctx, &parser, args)?;
-  drop(parser);
 
-  subargs.insert(0, format!("nitrocli {}", command));
+  let mut stdout_buf = BufWriter::new(ctx.stdout);
+  let mut stderr_buf = BufWriter::new(ctx.stderr);
+  let mut stdio_buf = (&mut stdout_buf, &mut stderr_buf);
+  let result = parse(&mut stdio_buf, parser, args);
 
-  let ctx = ExecCtx {
-    model,
-    stdout: ctx.stdout,
-    stderr: ctx.stderr,
-    admin_pin: ctx.admin_pin.take(),
-    user_pin: ctx.user_pin.take(),
-    new_admin_pin: ctx.new_admin_pin.take(),
-    new_user_pin: ctx.new_user_pin.take(),
-    password: ctx.password.take(),
-    verbosity,
-  };
-  Ok((command, ctx, subargs))
-}
+  if version {
+    println!(ctx, "nitrocli {}", env!("CARGO_PKG_VERSION"))?;
+    Ok(())
+  } else {
+    stdout_buf.flush()?;
+    stderr_buf.flush()?;
 
-/// Parse the command-line arguments and execute the selected command.
-pub(crate) fn handle_arguments(ctx: &mut RunCtx<'_>, args: Vec<String>) -> Result<()> {
-  let (command, mut ctx, args) = parse_arguments(ctx, args)?;
-  command.execute(&mut ctx, args)
+    result?;
+    subargs.insert(0, format!("nitrocli {}", command));
+
+    let mut ctx = ExecCtx {
+      model,
+      stdout: ctx.stdout,
+      stderr: ctx.stderr,
+      admin_pin: ctx.admin_pin.take(),
+      user_pin: ctx.user_pin.take(),
+      new_admin_pin: ctx.new_admin_pin.take(),
+      new_user_pin: ctx.new_user_pin.take(),
+      password: ctx.password.take(),
+      no_cache: ctx.no_cache,
+      verbosity,
+    };
+    command.execute(&mut ctx, subargs)
+  }
 }
