@@ -41,6 +41,7 @@ mod lock;
 mod otp;
 mod pin;
 mod pws;
+mod reset;
 mod run;
 mod status;
 mod storage;
@@ -110,6 +111,14 @@ impl Nitrocli {
     result
   }
 
+  pub fn admin_pin(&mut self, pin: impl Into<ffi::OsString>) {
+    self.admin_pin = Some(pin.into())
+  }
+
+  pub fn new_admin_pin(&mut self, pin: impl Into<ffi::OsString>) {
+    self.new_admin_pin = Some(pin.into())
+  }
+
   pub fn user_pin(&mut self, pin: impl Into<ffi::OsString>) {
     self.user_pin = Some(pin.into())
   }
@@ -148,6 +157,7 @@ impl Nitrocli {
       new_admin_pin: self.new_admin_pin.clone(),
       new_user_pin: self.new_user_pin.clone(),
       password: self.password.clone(),
+      no_cache: true,
     };
 
     (f(ctx, args), stdout, stderr)
